#!/bin/bash
#
# Use a single hooke domain with a constant unfolding rate, so
# unfolding force is proportional to time, and connect multiple
# identical constant unfolding domains.  We expect an exponential
# population decay.
#
# e.g.
#   spring constant k=df/dx      \__
#   velocity        v=dx/dt      /  `--> df/dt = kv --> f = kvt  (+ f0)
#   unfolding rate  K=1 frac/s
#
#   Population as a function of force/time follows
#     p(t) = exp(-tK) = exp(-FK/kv) = p(F)
#   so histogram of unfolding vs. force p(F) normalized to p(0)=1 should follow
#     -binwidth N0 dp/dF = binwidth N0 pK/kv = binwidth N0 K/kv exp(-FK/kv)
#
# usage: const_rate.sh <num_domains> <velocity> <spring_constant> <unfolding_rate>

if [ $# -ne 4 ]
then
    echo "usage: $0 <num_domains> <velocity> <spring_constant> <unfolding_rate>"
    exit 1
fi

NDOM=$1
V=$2
K=$3
RATE=$4
DEBUG=0

# since we're on the cluster, we'll go crazy and get really good statistics ;)
N=10000
df=`python -c "print 0.1/$RATE"`

SAWSIM=../../bin/sawsim
DATA=const_rate_${NDOM}_${V}_${K}_${RATE}.d
HIST=const_rate_${NDOM}_${V}_${K}_${RATE}.hist
COMPFILE=const_rate_${NDOM}_${V}_${K}_${RATE}.compare
GPFILE=const_rate_${NDOM}_${V}_${K}_${RATE}.gp
PNGFILE=const_rate_${NDOM}_${V}_${K}_${RATE}.png

# set up the theory
# our predicted histogram (see notes above) is
#  hist(F) = binwidth N0 K/kv exp(-FK/kv) = A exp(F/B)
theoryA=`python -c "print $df*$N*$NDOM*$RATE/float($K*$V)"`
theoryB=`python -c "print float(-$K*$V)/$RATE"`
THEORY_FN="$theoryA * exp(x / ($theoryB))"


# run the experiments
> $DATA
if [ "$ISACLUSTER" -eq 1 ]
then
    # Sawsim >= 0.6
    qcmd xtimes $N $SAWSIM -v$V -s cantilever,hooke,$K -N1 \
	-s folded,null -N$NDOM -s unfolded,null \
        -k folded,unfolded,const,$RATE -q folded \
        | grep -v '^#' | cut -f1 >> $DATA
else
    # Sawsim >= 0.6
    xtimes $N $SAWSIM -v$V -s cantilever,hooke,$K -N1 \
	-s folded,null -N$NDOM -s unfolded,null \
        -k folded,unfolded,const,$RATE -q folded \
        | grep -v '^#' | cut -f1 >> $DATA
fi

# histogram the data
if [ $DEBUG -eq 1 ]; then
    echo "cat $DATA | stem_leaf -v-1 -b$df -c | sed 's/://' > $HIST"
fi
cat $DATA | stem_leaf -v-1 -b$df -c | sed 's/://' > $HIST

# fit the data to an exponential decay
if [ $DEBUG -eq 1 ]; then
    echo "cat $HIST | ./fit_exponential"
fi
FIT=`cat $HIST | ./fit_exponential` || exit 1
B=`echo "$FIT" | sed -n 's/time constant:[[:space:]]*//p'`
A=`echo "$FIT" | sed -n 's/initial value:[[:space:]]*//p'`
echo -e "force scale:\t$B\t(expected $theoryB)" > $COMPFILE
echo -e "initial value:\t$A\t(expected $theoryA)" >> $COMPFILE
cat $COMPFILE

# generate gnuplot script
if [ $DEBUG -eq 1 ]; then
    echo "generate Gnuplot script"
fi
echo "set terminal png" > $GPFILE
echo "set output '$PNGFILE'" >> $GPFILE
echo "set logscale y" >> $GPFILE
echo "set xlabel 'Force'" >> $GPFILE
echo "set ylabel 'Deaths'" >> $GPFILE
echo "theory(x) = $THEORY_FN" >> $GPFILE
echo "fit(x) = $A * exp(x / $B)" >> $GPFILE
echo "plot  theory(x), '$HIST' with points, fit(x)" >> $GPFILE

if [ $DEBUG -eq 1 ]; then
    echo "Generate PNG"
fi
gnuplot $GPFILE
